#include "jabber/jlib.h"

thread xmlstream_new(int fd, xmlstream_onNode evt)
{
    debug("xmlstream_new");
    return tstream_new(fd, _xmlstream_main, evt);
}


int xmlstream_write(thread t, xmlnode x)
{
    char *buff;
    int len;

    debug("xmlstream_write");

    if(t == NULL || x == NULL)
        return 0;

    /* serialize xmlnode to a string */
    buff = xmlnode2str(x);
    tstream_write(t, buff, strlen(buff));

    return 0;
}

/* give a standard template xmlnode to work from */
xmlnode xmlstream_header(char *namespace, char *to, char *from)
{
    xmlnode x;

    x = xmlnode_new_tag("stream:stream");
    xmlnode_put_attrib(x, "xmlns:stream", "http://etherx.jabber.org/streams");
    if(namespace != NULL)
        xmlnode_put_attrib(x, "xmlns", namespace);
    if(to != NULL)
        xmlnode_put_attrib(x, "to", to);
    if(from != NULL)
        xmlnode_put_attrib(x, "from", from);

    return x;
}

/* special wrapper to only write the root opening */
void xmlstream_write_header(thread t, xmlnode x)
{
    char *head, *fixr;

    if(xmlnode_has_children(x))
        return;

    head = strgrow(NULL,"<?xml version=\"1.0\"?>",0,0);
    head = strgrow(head,xmlnode2str(x),0,0);
    fixr = strstr(head,"/>");
    *fixr = '>';
    ++fixr;
    *fixr = '\0';

    tstream_write(t, head, -1);
    free(head);
    xmlnode_free(x);
}

/* special wrapper to send an error and close this stream */
void xmlstream_error(thread t, char *err)
{
    tstream_write(t, "<stream:error>", 14);
    tstream_write(t, err, -1);
    tstream_write(t, "</stream:error></stream:stream>", 31);
    if(t->id == pth_self())
        pth_exit(NULL);
    else
        pth_cancel(t->id);
}

void _xmlstream_startElement(void* userdata, const char* name, const char** atts)
{
    thread t = (thread)userdata;
    tstream ts = (tstream)(t->data);
    xmlnode current = (xmlnode)(ts->data_aux2);
    xmlstream_onNode onNode = (xmlstream_onNode)(ts->read_evt_data);

    insane("stream_startElement: %s",name);

    if(current == NULL)
    {
        insane("stream_startElement: %s","creating new node");
        current = xmlnode_new_tag(name);
        xmlnode_put_expat_attribs(current, atts);

        if(ts->data_aux3 == NULL)
        { /* this is a special flag saying that this is the <stream> root */
            insane("stream_startElement: %s","processing root node");
            ts->data_aux3 = (void *)1;
            (onNode)(t, current, XMLSTREAM_ROOT);
            xmlnode_free(current);
            current = NULL;
        }
    }
    else
    {
        insane("startElement: %s","creating child node");
        current = xmlnode_insert_tag(current, name);
        xmlnode_put_expat_attribs(current, atts);
    }

    ts->data_aux2 = current;
}


void _xmlstream_endElement(void* userdata, const char* name)
{
    thread t = (thread)userdata;
    tstream ts = (tstream)(t->data);
    xmlnode current = (xmlnode)(ts->data_aux2);
    xmlstream_onNode onNode = (xmlstream_onNode)(ts->read_evt_data);
    xmlnode parent;

    insane("stream_endElement: %s",name);

    /* if it's already NULL we've received </stream>, tell the app and we're outta here */
    if(current == NULL)
    {
        (onNode)(t, current, XMLSTREAM_CLOSE);
        pth_exit(NULL);
    }else{
        parent = xmlnode_get_parent(current);
        if(parent == NULL)
        { /* we are the top-most node, feed to the app and delete */
            (onNode)(t, current, XMLSTREAM_NODE);
            xmlnode_free(current);
        }
        ts->data_aux2 = parent;
    }
}


void _xmlstream_charData(void* userdata, const char* s, int len)
{
    thread t = (thread)userdata;
    tstream ts = (tstream)(t->data);
    xmlnode current = (xmlnode)(ts->data_aux2);
    xmlstream_onNode onNode = (xmlstream_onNode)(ts->read_evt_data);

    if(current == NULL)
    {
        /* we must be in the root of the stream where CDATA is irrelevant */
        return;
    }

    insane("_xmlstream_charData: %s",s);
    xmlnode_insert_cdata(current, s, len);
}


void _xmlstream_main(void *arg, void *buffer, int bufsz)
{
    /*
    this is called when there is new data on the tstream
    it should process the XML using the special stream expat callbacks which
    call (ts->read_evt_data)()
    */

    thread t = (thread)arg;
    tstream ts = (tstream)(t->data);
    xmlstream_onNode onNode = (xmlstream_onNode)ts->read_evt_data;
    XML_Parser p = (XML_Parser)(ts->data_aux1);

    debug("_xmlstream_main");

    if(buffer == NULL)
    { /* init */
        insane("_xmlstream_main: init");
        pth_cleanup_push(_xmlstream_cleanup, (void *)t);
        p = XML_ParserCreate(NULL);
        ts->data_aux1 = (void *)p;
        XML_SetUserData(p, (void *)t);
        XML_SetElementHandler(p, (void *)_xmlstream_startElement, (void *)_xmlstream_endElement);
        XML_SetCharacterDataHandler(p, (void *)_xmlstream_charData);
        (onNode)(t, NULL, XMLSTREAM_INIT);
        return;
    }

    insane("XML_Parse: %s",buffer);

    if(!XML_Parse(p, buffer, bufsz, 0))
    {
        error("xmlstream","xmlstream_parse_error: %s",(char *)XML_ErrorString(XML_GetErrorCode(p)));
        (onNode)(t, NULL, XMLSTREAM_ERROR);
        pth_exit(NULL);
    }
}

void _xmlstream_cleanup(void *arg)
{
    thread t = (thread)arg;
    tstream ts = (tstream)(t->data);
    xmlnode *x = (xmlnode *)(ts->data_aux2);
    XML_Parser p = (XML_Parser)(ts->data_aux1);

    debug("_xmlstream_cleanup");
    XML_ParserFree(p);
}
